/* SPDX-License-Identifier: GPL-2.0 */
/*
 * cpu.h: Values of the PRId register used to match up
 *	  various LoongArch cpu types.
 *
 * Copyright (C) 1996 David S. Miller (davem@davemloft.net)
 * Copyright (C) 2004, 2013  Maciej W. Rozycki
 * Copyright (C) 2020 Loongson Technology Corporation Limited
 */
#ifndef _ASM_CPU_H
#define _ASM_CPU_H

/*
   As of the LoongArch specs from Loongson Technology, the PRId register
   (CPUCFG.00) is defined in this (backwards compatible) way:

  +----------------+----------------+----------------+----------------+
  | Company Options| Company ID	    | Processor ID   | Revision	      |
  +----------------+----------------+----------------+----------------+
   31		 24 23		  16 15		    8 7

*/

#define PRID_OPT_MASK		0xff000000

/*
 * Assigned Company values for bits 23:16 of the PRId register.
 */

#define PRID_COMP_MASK		0xff0000

#define PRID_COMP_LOONGSON	0x140000

/*
 * Assigned Processor ID (implementation) values for bits 15:8 of the PRId
 * register.  In order to detect a certain CPU type exactly eventually
 * additional registers may need to be examined.
 */

#define PRID_IMP_MASK		0xff00

#define PRID_IMP_LOONGSON_32	0x4200  /* Loongson 32bit */
#define PRID_IMP_LOONGSON_64R	0x6100  /* Reduced Loongson 64bit */
#define PRID_IMP_LOONGSON_64C	0x6300  /* Classic Loongson 64bit */
#define PRID_IMP_LOONGSON_64G	0xc000  /* Generic Loongson 64bit */
#define PRID_IMP_UNKNOWN	0xff00

/*
 * Particular Revision values for bits 7:0 of the PRId register.
 */

#define PRID_REV_MASK		0x00ff

#if !defined(__ASSEMBLY__)

enum cpu_type_enum {
	CPU_UNKNOWN,
	CPU_LOONGSON32,
	CPU_LOONGSON64,
	CPU_LAST
};

#endif /* !__ASSEMBLY */

/*
 * ISA Level encodings
 *
 */
#define LOONGARCH_CPU_ISA_LA32R 0x00000001
#define LOONGARCH_CPU_ISA_LA32S 0x00000002
#define LOONGARCH_CPU_ISA_LA64  0x00000004

#define LOONGARCH_CPU_ISA_32BIT (LOONGARCH_CPU_ISA_LA32R | LOONGARCH_CPU_ISA_LA32S)
#define LOONGARCH_CPU_ISA_64BIT LOONGARCH_CPU_ISA_LA64

/*
 * CPU Option encodings
 */
#define LOONGARCH_CPU_TLB		BIT_ULL(0)	/* CPU has TLB */
#define LOONGARCH_CPU_CSR		BIT_ULL(1)	/* CPU has CSR feature */
#define LOONGARCH_CPU_FPU		BIT_ULL(2)	/* CPU has FPU */
#define LOONGARCH_CPU_LLSC		BIT_ULL(3)	/* CPU has ll/sc instructions */
#define LOONGARCH_CPU_WATCH		BIT_ULL(4)	/* watchpoint registers */
#define LOONGARCH_CPU_VINT		BIT_ULL(5)	/* CPU supports vectored interrupts */
#define LOONGARCH_CPU_EXTIOI		BIT_ULL(6)	/* CPU has EXTIOI */
#define LOONGARCH_CPU_CSRIPI		BIT_ULL(7)	/* CPU has CSR IPI */
#define LOONGARCH_CPU_PREFETCH		BIT_ULL(8)	/* CPU has prefetch instructions */
#define LOONGARCH_CPU_LPA		BIT_ULL(9)	/* CPU supports Large Physical Addressing */
#define LOONGARCH_CPU_LFT		BIT_ULL(10)	/* CPU has Local Fix Timer feature */
#define LOONGARCH_CPU_GFT		BIT_ULL(11)	/* CPU has Global Fix Timer feature */
#define LOONGARCH_CPU_PMP		BIT_ULL(12)	/* CPU has perfermance counter */
#define LOONGARCH_CPU_SCALEFREQ		BIT_ULL(13)	/* CPU support scale cpufreq */
#define LOONGARCH_CPU_GUESTID		BIT_ULL(14)	/* CPU has VZ GuestCtl0Ext register */
#define LOONGARCH_CPU_HYPERVISOR	BIT_ULL(15)	/* CPU has VZ GuestCtl1 register */

/*
 * CPU ASE encodings
 */
#define LOONGARCH_ASE_CPUCFG		0x00000001 /* CPU supports CPUCFG */
#define LOONGARCH_ASE_LAMO		0x00000002 /* CPU supports Atomic instructions */
#define LOONGARCH_ASE_LSX		0x00000004 /* CPU supports 128bit SIMD instructions */
#define LOONGARCH_ASE_LASX		0x00000008 /* CPU supports 256bit SIMD instructions */
#define LOONGARCH_ASE_LVZ		0x00000010 /* CPU supports Virtualization ASE */
#define LOONGARCH_ASE_LBT		0x00000020 /* CPU supports Binary Translation ASE */
#define LOONGARCH_ASE_CAM		0x00000040 /* CPU supports CAM feature */

#endif /* _ASM_CPU_H */
