/* SPDX-License-Identifier: GPL-2.0 */
/*
* This file is subject to the terms and conditions of the GNU General Public
* License.  See the file "COPYING" in the main directory of this archive
* for more details.
*
* Copyright (C) 2020 Loongson Technologies, Inc.  All rights reserved.
* Authors: Xing Li <lixing@loongson.cn>
*/

#ifndef __LOONGARCH_KVM_HOST_H__
#define __LOONGARCH_KVM_HOST_H__

#include <linux/cpumask.h>
#include <linux/mutex.h>
#include <linux/hrtimer.h>
#include <linux/interrupt.h>
#include <linux/types.h>
#include <linux/kvm.h>
#include <linux/kvm_types.h>
#include <linux/threads.h>
#include <linux/spinlock.h>

#include <asm/inst.h>
#include <asm/loongarchregs.h>
#include <asm/compiler.h>

/* Loongarch KVM register ids */
#define LOONGARCH_CSR_32(_R, _S)					\
	(KVM_REG_LOONGARCH_CSR | KVM_REG_SIZE_U32 | (8 * (_R) + (_S)))

#define LOONGARCH_CSR_64(_R, _S)					\
	(KVM_REG_LOONGARCH_CSR | KVM_REG_SIZE_U64 | (8 * (_R) + (_S)))

#define KVM_LOONGARCH_CSR_CRMD		0
#define KVM_LOONGARCH_CSR_PRMD		1
#define KVM_LOONGARCH_CSR_EUEN		2
#define KVM_LOONGARCH_CSR_MISC		3
#define KVM_LOONGARCH_CSR_ECFG		4
#define KVM_LOONGARCH_CSR_ESTAT		5
#define KVM_LOONGARCH_CSR_EPC		6
#define KVM_LOONGARCH_CSR_BADV		7
#define KVM_LOONGARCH_CSR_BADI		8
#define KVM_LOONGARCH_CSR_EBASE		0xc
#define KVM_LOONGARCH_CSR_TLBIDX	0x10
#define KVM_LOONGARCH_CSR_TLBHI		0x11
#define KVM_LOONGARCH_CSR_TLBLO0	0x12
#define KVM_LOONGARCH_CSR_TLBLO1	0x13
#define KVM_LOONGARCH_CSR_GTLBC		0x15
#define KVM_LOONGARCH_CSR_TRGP		0x16
#define KVM_LOONGARCH_CSR_ASID		0x18
#define KVM_LOONGARCH_CSR_PGDL		0x19
#define KVM_LOONGARCH_CSR_PGDH		0x1a
#define KVM_LOONGARCH_CSR_PGD		0x1b
#define KVM_LOONGARCH_CSR_PWCTL0	0x1c
#define KVM_LOONGARCH_CSR_PWCTL1	0x1d
#define KVM_LOONGARCH_CSR_STLBPS	0x1e
#define KVM_LOONGARCH_CSR_RVACFG	0x1f
#define KVM_LOONGARCH_CSR_CPUNUM	0x20
#define KVM_LOONGARCH_CSR_PRCFG1	0x21
#define KVM_LOONGARCH_CSR_PRCFG2	0x22
#define KVM_LOONGARCH_CSR_PRCFG3	0x23
#define KVM_LOONGARCH_CSR_KSCRATCH0	0x30
#define KVM_LOONGARCH_CSR_KSCRATCH1	0x31
#define KVM_LOONGARCH_CSR_KSCRATCH2	0x32
#define KVM_LOONGARCH_CSR_KSCRATCH3	0x33
#define KVM_LOONGARCH_CSR_KSCRATCH4	0x34
#define KVM_LOONGARCH_CSR_KSCRATCH5	0x35
#define KVM_LOONGARCH_CSR_KSCRATCH6	0x36
#define KVM_LOONGARCH_CSR_KSCRATCH7	0x37
#define KVM_LOONGARCH_CSR_KSCRATCH8	0x38
#define KVM_LOONGARCH_CSR_TIMERID	0x40
#define KVM_LOONGARCH_CSR_TIMERCFG	0x41
#define KVM_LOONGARCH_CSR_TIMERTICK	0x42
#define KVM_LOONGARCH_CSR_TIMEROFFSET	0x43
#define KVM_LOONGARCH_CSR_GSTAT		0x50
#define KVM_LOONGARCH_CSR_GCFG		0x51
#define KVM_LOONGARCH_CSR_GINTC		0x52
#define KVM_LOONGARCH_CSR_GCNTC		0x53
#define KVM_LOONGARCH_CSR_LLBCTL	0x60
#define KVM_LOONGARCH_CSR_IMPCTL1	0x80
#define KVM_LOONGARCH_CSR_IMPCTL2	0x81
#define KVM_LOONGARCH_CSR_GNMI		0x82
#define KVM_LOONGARCH_CSR_TLBREBASE	0x88
#define KVM_LOONGARCH_CSR_TLBRBADV	0x89
#define KVM_LOONGARCH_CSR_TLBREPC	0x8a
#define KVM_LOONGARCH_CSR_TLBRSAVE	0x8b
#define KVM_LOONGARCH_CSR_TLBRELO0	0x8c
#define KVM_LOONGARCH_CSR_TLBRELO1	0x8d
#define KVM_LOONGARCH_CSR_TLBREHI	0x8e
#define KVM_LOONGARCH_CSR_TLBRPRMD	0x8f
#define KVM_LOONGARCH_CSR_ERRCTL	0x90
#define KVM_LOONGARCH_CSR_ERRINFO1	0x91
#define KVM_LOONGARCH_CSR_ERRINFO2	0x92
#define KVM_LOONGARCH_CSR_ERREBASE	0x93
#define KVM_LOONGARCH_CSR_ERREPC	0x94
#define KVM_LOONGARCH_CSR_ERRSAVE	0x95
#define KVM_LOONGARCH_CSR_CTAG		0x98
#define KVM_LOONGARCH_CSR_MCSR0		0xc0
#define KVM_LOONGARCH_CSR_MCSR1		0xc1
#define KVM_LOONGARCH_CSR_MCSR2		0xc2
#define KVM_LOONGARCH_CSR_MCSR3		0xc3
#define KVM_LOONGARCH_CSR_MCSR8		0xc8
#define KVM_LOONGARCH_CSR_MCSR9		0xc9
#define KVM_LOONGARCH_CSR_MCSR10	0xca
#define KVM_LOONGARCH_CSR_MCSR24	0xf0
#define KVM_LOONGARCH_CSR_UCAWIN	0x100
#define KVM_LOONGARCH_CSR_UCAWIN0_LO	0x102
#define KVM_LOONGARCH_CSR_UCAWIN0_HI	0x103
#define KVM_LOONGARCH_CSR_UCAWIN1_LO	0x104
#define KVM_LOONGARCH_CSR_UCAWIN1_HI	0x105
#define KVM_LOONGARCH_CSR_UCAWIN2_LO	0x106
#define KVM_LOONGARCH_CSR_UCAWIN2_HI	0x107
#define KVM_LOONGARCH_CSR_UCAWIN3_LO	0x108
#define KVM_LOONGARCH_CSR_UCAWIN3_HI	0x109
#define KVM_LOONGARCH_CSR_DMWIN0	0x180
#define KVM_LOONGARCH_CSR_DMWIN1	0x181
#define KVM_LOONGARCH_CSR_DMWIN2	0x182
#define KVM_LOONGARCH_CSR_DMWIN3	0x183
#define KVM_LOONGARCH_CSR_PERF0_EVENT	0x200
#define KVM_LOONGARCH_CSR_PERF0_COUNT	0x201
#define KVM_LOONGARCH_CSR_PERF1_EVENT	0x202
#define KVM_LOONGARCH_CSR_PERF1_COUNT	0x203
#define KVM_LOONGARCH_CSR_PERF2_EVENT	0x204
#define KVM_LOONGARCH_CSR_PERF2_COUNT	0x205
#define KVM_LOONGARCH_CSR_PERF3_EVENT	0x206
#define KVM_LOONGARCH_CSR_PERF3_COUNT	0x207
#define KVM_LOONGARCH_CSR_DEBUG		0x500
#define KVM_LOONGARCH_CSR_DEPC		0x501
#define KVM_LOONGARCH_CSR_DESAVE	0x502

#define KVM_CSR_CRMD		LOONGARCH_CSR_64(0, 0)
#define KVM_CSR_PRMD		LOONGARCH_CSR_64(1, 0)
#define KVM_CSR_EUEN		LOONGARCH_CSR_64(2, 0)
#define KVM_CSR_MISC		LOONGARCH_CSR_64(3, 0)
#define KVM_CSR_ECFG		LOONGARCH_CSR_64(4, 0)
#define KVM_CSR_ESTAT		LOONGARCH_CSR_64(5, 0)
#define KVM_CSR_EPC		LOONGARCH_CSR_64(6, 0)
#define KVM_CSR_BADV		LOONGARCH_CSR_64(7, 0)
#define KVM_CSR_BADI		LOONGARCH_CSR_64(8, 0)
#define KVM_CSR_EBASE		LOONGARCH_CSR_64(0xc, 0)
#define KVM_CSR_TLBIDX		LOONGARCH_CSR_64(0x10, 0)
#define KVM_CSR_TLBHI		LOONGARCH_CSR_64(0x11, 0)
#define KVM_CSR_TLBLO0		LOONGARCH_CSR_64(0x12, 0)
#define KVM_CSR_TLBLO1		LOONGARCH_CSR_64(0x13, 0)
#define KVM_CSR_GTLBC		LOONGARCH_CSR_64(0x15, 0)
#define KVM_CSR_TRGP		LOONGARCH_CSR_64(0x16, 0)
#define KVM_CSR_ASID		LOONGARCH_CSR_64(0x18, 0)
#define KVM_CSR_PGDL		LOONGARCH_CSR_64(0x19, 0)
#define KVM_CSR_PGDH		LOONGARCH_CSR_64(0x1a, 0)
#define KVM_CSR_PGD		LOONGARCH_CSR_64(0x1b, 0)
#define KVM_CSR_PWCTL0		LOONGARCH_CSR_64(0x1c, 0)
#define KVM_CSR_PWCTL1		LOONGARCH_CSR_64(0x1d, 0)
#define KVM_CSR_STLBPS		LOONGARCH_CSR_64(0x1e, 0)
#define KVM_CSR_RVACFG		LOONGARCH_CSR_64(0x1f, 0)
#define KVM_CSR_CPUNUM		LOONGARCH_CSR_64(0x20, 0)
#define KVM_CSR_PRCFG1		LOONGARCH_CSR_64(0x21, 0)
#define KVM_CSR_PRCFG2		LOONGARCH_CSR_64(0x22, 0)
#define KVM_CSR_PRCFG3		LOONGARCH_CSR_64(0x23, 0)
#define KVM_CSR_KSCRATCH0	LOONGARCH_CSR_64(0x30, 0)
#define KVM_CSR_KSCRATCH1	LOONGARCH_CSR_64(0x31, 0)
#define KVM_CSR_KSCRATCH2	LOONGARCH_CSR_64(0x32, 0)
#define KVM_CSR_KSCRATCH3	LOONGARCH_CSR_64(0x33, 0)
#define KVM_CSR_KSCRATCH4	LOONGARCH_CSR_64(0x34, 0)
#define KVM_CSR_KSCRATCH5	LOONGARCH_CSR_64(0x35, 0)
#define KVM_CSR_KSCRATCH6	LOONGARCH_CSR_64(0x36, 0)
#define KVM_CSR_KSCRATCH7	LOONGARCH_CSR_64(0x37, 0)
#define KVM_CSR_KSCRATCH8	LOONGARCH_CSR_64(0x38, 0)
#define KVM_CSR_TIMERID		LOONGARCH_CSR_64(0x40, 0)
#define KVM_CSR_TIMERCFG	LOONGARCH_CSR_64(0x41, 0)
#define KVM_CSR_TIMERTICK	LOONGARCH_CSR_64(0x42, 0)
#define KVM_CSR_TIMEROFFSET	LOONGARCH_CSR_64(0x43, 0)
#define KVM_CSR_GSTAT		LOONGARCH_CSR_64(0x50, 0)
#define KVM_CSR_GCFG		LOONGARCH_CSR_64(0x51, 0)
#define KVM_CSR_GINTC		LOONGARCH_CSR_64(0x52, 0)
#define KVM_CSR_GCNTC		LOONGARCH_CSR_64(0x53, 0)
#define KVM_CSR_LLBCTL		LOONGARCH_CSR_64(0x60, 0)
#define KVM_CSR_IMPCTL1		LOONGARCH_CSR_64(0x80, 0)
#define KVM_CSR_IMPCTL2		LOONGARCH_CSR_64(0x81, 0)
#define KVM_CSR_GNMI		LOONGARCH_CSR_64(0x82, 0)
#define KVM_CSR_TLBREBASE	LOONGARCH_CSR_64(0x88, 0)
#define KVM_CSR_TLBRBADV	LOONGARCH_CSR_64(0x89, 0)
#define KVM_CSR_TLBREPC		LOONGARCH_CSR_64(0x8a, 0)
#define KVM_CSR_TLBRSAVE	LOONGARCH_CSR_64(0x8b, 0)
#define KVM_CSR_TLBRELO0	LOONGARCH_CSR_64(0x8c, 0)
#define KVM_CSR_TLBRELO1	LOONGARCH_CSR_64(0x8d, 0)
#define KVM_CSR_TLBREHI		LOONGARCH_CSR_64(0x8e, 0)
#define KVM_CSR_TLBRPRMD	LOONGARCH_CSR_64(0x8f, 0)
#define KVM_CSR_ERRCTL		LOONGARCH_CSR_64(0x90, 0)
#define KVM_CSR_ERRINFO1	LOONGARCH_CSR_64(0x91, 0)
#define KVM_CSR_ERRINFO2	LOONGARCH_CSR_64(0x92, 0)
#define KVM_CSR_ERREBASE	LOONGARCH_CSR_64(0x93, 0)
#define KVM_CSR_ERREPC		LOONGARCH_CSR_64(0x94, 0)
#define KVM_CSR_ERRSAVE		LOONGARCH_CSR_64(0x95, 0)
#define KVM_CSR_CTAG		LOONGARCH_CSR_64(0x98, 0)
#define KVM_CSR_MCSR0		LOONGARCH_CSR_64(0xc0, 0)
#define KVM_CSR_MCSR1		LOONGARCH_CSR_64(0xc1, 0)
#define KVM_CSR_MCSR2		LOONGARCH_CSR_64(0xc2, 0)
#define KVM_CSR_MCSR3		LOONGARCH_CSR_64(0xc3, 0)
#define KVM_CSR_MCSR8		LOONGARCH_CSR_64(0xc8, 0)
#define KVM_CSR_MCSR9		LOONGARCH_CSR_64(0xc9, 0)
#define KVM_CSR_MCSR10		LOONGARCH_CSR_64(0xca, 0)
#define KVM_CSR_MCSR24		LOONGARCH_CSR_64(0xf0, 0)
#define KVM_CSR_UCWIN		LOONGARCH_CSR_64(0x100, 0)
#define KVM_CSR_UCWIN0_LO	LOONGARCH_CSR_64(0x102, 0)
#define KVM_CSR_UCWIN0_HI	LOONGARCH_CSR_64(0x103, 0)
#define KVM_CSR_UCWIN1_LO	LOONGARCH_CSR_64(0x104, 0)
#define KVM_CSR_UCWIN1_HI	LOONGARCH_CSR_64(0x105, 0)
#define KVM_CSR_UCWIN2_LO	LOONGARCH_CSR_64(0x106, 0)
#define KVM_CSR_UCWIN2_HI	LOONGARCH_CSR_64(0x107, 0)
#define KVM_CSR_UCWIN3_LO	LOONGARCH_CSR_64(0x108, 0)
#define KVM_CSR_UCWIN3_HI	LOONGARCH_CSR_64(0x109, 0)
#define KVM_CSR_DMWIN0		LOONGARCH_CSR_64(0x180, 0)
#define KVM_CSR_DMWIN1		LOONGARCH_CSR_64(0x181, 0)
#define KVM_CSR_DMWIN2		LOONGARCH_CSR_64(0x182, 0)
#define KVM_CSR_DMWIN3		LOONGARCH_CSR_64(0x183, 0)
#define KVM_CSR_PERF0_EVENT	LOONGARCH_CSR_64(0x200, 0)
#define KVM_CSR_PERF0_COUNT	LOONGARCH_CSR_64(0x201, 0)
#define KVM_CSR_PERF1_EVENT	LOONGARCH_CSR_64(0x202, 0)
#define KVM_CSR_PERF1_COUNT	LOONGARCH_CSR_64(0x203, 0)
#define KVM_CSR_PERF2_EVENT	LOONGARCH_CSR_64(0x204, 0)
#define KVM_CSR_PERF2_COUNT	LOONGARCH_CSR_64(0x205, 0)
#define KVM_CSR_PERF3_EVENT	LOONGARCH_CSR_64(0x206, 0)
#define KVM_CSR_PERF3_COUNT	LOONGARCH_CSR_64(0x207, 0)
#define KVM_CSR_DEBUG		LOONGARCH_CSR_64(0x500, 0)
#define KVM_CSR_DEPC		LOONGARCH_CSR_64(0x501, 0)
#define KVM_CSR_DESAVE		LOONGARCH_CSR_64(0x502, 0)

#define KVM_MAX_VCPUS		16
#define KVM_USER_MEM_SLOTS	16
/* memory slots that does not exposed to userspace */
#define KVM_PRIVATE_MEM_SLOTS	0

#define KVM_HALT_POLL_NS_DEFAULT 500000

extern unsigned long GUESTID_MASK;
extern unsigned long GUESTID_FIRST_VERSION;
extern unsigned long GUESTID_VERSION_MASK;

typedef union loongarch_instruction  larch_inst;

#define KVM_INVALID_ADDR		0xdeadbeef
#define KVM_HVA_ERR_BAD			(-1UL)
#define KVM_HVA_ERR_RO_BAD		(-2UL)
static inline bool kvm_is_error_hva(unsigned long addr)
{
	return IS_ERR_VALUE(addr);
}

struct kvm_vm_stat {
	ulong remote_tlb_flush;
	u64 lvz_kvm_vm_ioctl_irq_line;
	u64 lvz_kvm_ls7a_ioapic_update;
	u64 lvz_kvm_ls7a_ioapic_set_irq;
	u64 lvz_ls7a_ioapic_reg_write;
	u64 lvz_ls7a_ioapic_reg_read;
	u64 lvz_kvm_set_ls7a_ioapic;
	u64 lvz_kvm_get_ls7a_ioapic;
	u64 lvz_kvm_set_ls3a_ht_irq;
	u64 lvz_kvm_get_ls3a_ht_irq;
	u64 lvz_kvm_set_ls3a_router_irq;
	u64 lvz_kvm_get_ls3a_router_irq;
	u64 lvz_kvm_set_ls3a_ext_irq;
	u64 lvz_kvm_get_ls3a_ext_irq;
	u64 lvz_kvm_set_ls3a_ipmask;
	u64 lvz_kvm_get_ls3a_ipmask;
	u64 lvz_kvm_pip_read_exits;
	u64 lvz_kvm_pip_write_exits;
	u64 lvz_kvm_ls7a_pic_read_exits;
	u64 lvz_kvm_ls7a_pic_write_exits;
	u64 lvz_kvm_ls3a_ht_write_exits;
	u64 lvz_kvm_ls3a_ht_read_exits;
	u64 lvz_kvm_ls3a_router_write_exits;
	u64 lvz_kvm_ls3a_router_read_exits;
	u64 lvz_kvm_ls7a_msi_irq;
};

struct kvm_vcpu_stat {
	u64 wait_exits;
	u64 cache_exits;
	u64 signal_exits;
	u64 int_exits;
	u64 tlbmiss_ld_exits;
	u64 tlbmiss_st_exits;
	u64 tlbmiss_ifetch_exits;
	u64 tlbmod_exits;
	u64 tlbri_exits;
	u64 tlbxi_exits;
	u64 tlbprev_exits;
	u64 addrerr_exits;
	u64 unalign_exits;
	u64 bound_exits;
	u64 syscall_exits;
	u64 break_inst_exits;
	u64 resvd_inst_exits;
	u64 prev_inst_exits;
	u64 trap_inst_exits;
	u64 fp_disabled_exits;
	u64 lsx_disabled_exits;
	u64 lasx_disabled_exits;
	u64 fpe_exits;
	u64 watch_exits;
	u64 lvz_gpsi_exits;
	u64 lvz_hc_exits;
	u64 lvz_gfc_exits;
	u64 flush_dcache_exits;
	u64 lvz_mmio_exits;
	u64 lvz_general_exits;
	u64 lvz_ignore_exits;
	u64 lvz_nc_exits;
	u64 lvz_hc_tlbmiss_exits;
	u64 lvz_hc_tlbm_exits;
	u64 lvz_hc_tlbl_exits;
	u64 lvz_hc_tlbs_exits;
	u64 lvz_hc_emulate_exits;
	u64 lvz_hc_missvalid_exits;
	u64 lvz_gpsi_csr_exits;
	u64 lvz_gpsi_cache_exits;
	u64 lvz_tlb_refill_exits;
	u64 lvz_tlb_refill_fail;
	u64 lvz_tlb_add_replace;
	u64 lvz_tlb_add_refresh;
	u64 lvz_rdcsr_exits;
	u64 lvz_rdcsr_cpu_feature_exits;
	u64 lvz_rdcsr_cpu_temperature_exits;
	u64 lvz_rdcsr_node_counter_exits;
	u64 lvz_rdcsr_misc_func_exits;
	u64 lvz_rdcsr_ipi_access_exits;
	u64 lvz_wrcsr_exits;
	u64 lvz_drdcsr_exits;
	u64 lvz_dwrcsr_exits;
	u64 lvz_cpucfg_exits;
	u64 lvz_stable_counter_exits;
	u64 lvz_huge_dec_exits;
	u64 lvz_huge_thp_exits;
	u64 lvz_huge_adjust_exits;
	u64 lvz_huge_set_exits;
	u64 lvz_huge_merge_exits;
	u64 halt_successful_poll;
	u64 halt_attempted_poll;
	u64 halt_poll_invalid;
	u64 halt_wakeup;
};

struct kvm_arch_memory_slot {
};

struct mailbox_val {
	union {
		struct {
			u32 lo;
			u32 hi;
		};
		u64 mailval;
	};
	u32 flags;
};

struct kvm_arch {
	/* Guest physical mm */
	struct mm_struct gpa_mm;
	/* Mask of CPUs needing GPA ASID flush */
	cpumask_t asid_flush_mask;

	unsigned char online_vcpus;
	unsigned char is_migrate;
	s64 nodecounter_offset;
	s64 stablecounter_gftoffset;
	u32 node_shift;
	u32 cpucfg_lasx;
	/* Check hypcall work or not */
	u32 hypcall_check;
	struct mailbox_val mval[NR_CPUS][4];
	struct ls7a_kvm_ioapic *v_ioapic;
	struct ls3a_kvm_ipi *v_gipi;
	struct ls3a_kvm_htirq *v_htirq;
	struct ls3a_kvm_routerirq *v_routerirq;
	struct ls3a_kvm_extirq *v_extirq;
	uint16_t core_ip_mask[KVM_MAX_VCPUS][4];
};


#define LOONGARCH_CSRS	0x100
#define CSR_UCWIN_BASE	0x100
#define CSR_UCWIN_SIZE	0x10
#define CSR_DMWIN_BASE	0x180
#define CSR_DMWIN_SIZE	0x4
#define CSR_PERF_BASE	0x200
#define CSR_PERF_SIZE	0x8
#define CSR_DEBUG_BASE	0x500
#define CSR_DEBUG_SIZE	0x3

struct loongarch_csrs {
	unsigned long csrs[LOONGARCH_CSRS];
	unsigned long csr_uca[CSR_UCWIN_SIZE];
	unsigned long csr_dmwin[CSR_DMWIN_SIZE];
	unsigned long csr_perf[CSR_PERF_SIZE];
	unsigned long csr_dbg[CSR_DEBUG_SIZE];
#ifdef CONFIG_KVM_LOONGARCH_DEBUG_COP0_COUNTERS
	unsigned long stat[LOONGARCH_CSRS];
#endif
};

/* Resume Flags */
#define RESUME_FLAG_DR		(1<<0)	/* Reload guest nonvolatile state? */
#define RESUME_FLAG_HOST	(1<<1)	/* Resume host? */

#define RESUME_GUEST		0
#define RESUME_GUEST_DR		RESUME_FLAG_DR
#define RESUME_HOST		RESUME_FLAG_HOST

enum emulation_result {
	EMULATE_DONE,		/* no further processing */
	EMULATE_DO_MMIO,	/* kvm_run filled with MMIO request */
	EMULATE_FAIL,		/* can't emulate this instruction */
	EMULATE_WAIT,		/* WAIT instruction */
	EMULATE_PRIV_FAIL,
	EMULATE_EXCEPT,		/* A guest exception has been generated */
	EMULATE_HYPERCALL,	/* HYPCALL instruction */
	EMULATE_DEBUG,		/* Emulate guest kernel debug */
};

#define loongarch3_paddr_to_tlbpfn(x) \
	(((unsigned long)(x) >> LOONGARCH3_PG_SHIFT) & LOONGARCH3_PG_FRAME)
#define loongarch3_tlbpfn_to_paddr(x) \
	((unsigned long)((x) & LOONGARCH3_PG_FRAME) << LOONGARCH3_PG_SHIFT)

#define LOONGARCH3_PG_SHIFT		6
#define LOONGARCH3_PG_FRAME		0x3fffffc0

#define VPN2_MASK		GENMASK(cpu_vmbits - 1, 12)
#define KVM_ENTRYHI_ASID	cpu_asid_mask(&current_cpu_data)
#define TLB_IS_GLOBAL(x)	((x).tlb_lo[0] & (x).tlb_lo[1] & ENTRYLO_G)
#define TLB_VPN2(x)		((x).tlb_hi & VPN2_MASK)
#define TLB_ASID(x)		((x).tlb_hi & KVM_ENTRYHI_ASID)
#define TLB_LO_IDX(x, va)	(((va) >> PAGE_SHIFT) & 1)
#define TLB_IS_VALID(x, va)	((x).tlb_lo[TLB_LO_IDX(x, va)] & ENTRYLO_V)
#define TLB_IS_DIRTY(x, va)	((x).tlb_lo[TLB_LO_IDX(x, va)] & ENTRYLO_D)
#define TLB_HI_ASID_HIT(x, y)	(TLB_IS_GLOBAL(x) ||			\
				 TLB_ASID(x) == ((y) & KVM_ENTRYHI_ASID))

struct kvm_loongarch_tlb {
	long tlb_asid;
	long tlb_idx;
	long tlb_hi;
	long tlb_lo[2];
};

#define KVM_NR_MEM_OBJS     4

/*
 * We don't want allocation failures within the mmu code, so we preallocate
 * enough memory for a single page fault in a cache.
 */
struct kvm_mmu_memory_cache {
	int nobjs;
	void *objects[KVM_NR_MEM_OBJS];
};

#if defined(CONFIG_CPU_HAS_LASX)
#define FPU_ALIGN		__aligned(32)
#elif defined(CONFIG_CPU_HAS_LSX)
#define FPU_ALIGN		__aligned(16)
#else
#define FPU_ALIGN
#endif
#define KVM_LOONGARCH_AUX_FPU	0x1
#define KVM_LOONGARCH_AUX_LSX	0x2
#define KVM_LOONGARCH_AUX_LASX	0x4

struct kvm_vcpu_arch {
	unsigned long guest_ebase;
	unsigned long host_ebase;
	int (*vcpu_run)(struct kvm_run *run, struct kvm_vcpu *vcpu);

	/* Host registers preserved across guest mode execution */
	unsigned long host_stack;
	unsigned long host_gp;
	unsigned long host_pgd;
	unsigned long host_pgdhi;
	unsigned long host_entryhi;

	/* Host CSR registers used when handling exits from guest */
	unsigned long host_badvaddr;
	unsigned long host_epc;
	unsigned long host_gtlbc;
	unsigned long host_estat;
	unsigned long host_crmd;
	unsigned long host_prmd;
	unsigned long host_badinstr;
	unsigned long host_gstat;
	unsigned long host_gcfg;
	unsigned long host_ecfg;

	u32 is_hypcall;
	/* GPRS */
	unsigned long gprs[32];
	unsigned long pc;

	/* FPU State */
	struct loongarch_fpu fpu FPU_ALIGN;
	/* Which auxiliary state is loaded (KVM_LOONGARCH_AUX_*) */
	unsigned int aux_inuse;

	/* CSR State */
	struct loongarch_csrs *csr;

	/* Resume PC after MMIO completion */
	unsigned long io_pc;
	/* GPR used as IO source/target */
	u32 io_gpr;

	struct hrtimer comparecount_timer;
	/* Count timer control KVM register */
	u32 count_ctl;

	/* Bitmask of exceptions that are pending */
	unsigned long pending_exceptions;

	/* Bitmask of pending exceptions to be cleared */
	unsigned long pending_exceptions_clr;

	/* Guest kernel/user [partial] mm */
	struct mm_struct guest_kernel_mm, guest_user_mm;

	/* Guest ASID of last user mode execution */
	unsigned int last_user_gasid;

	/* Cache some mmu pages needed inside spinlock regions */
	struct kvm_mmu_memory_cache mmu_page_cache;

	/* vcpu's vzguestid is different on each host cpu in an smp system */
	u64 vzguestid[NR_CPUS];

	u64 stable_timer_tick;
	u64 stable_timer_cfg;
	/* Period of stable timer tick in ns */
	u64 stable_timer_period;
	/* Frequency of stable timer in Hz */
	u64 stable_timer_hz;
	/* Stable bias from the raw time */
	u64 stable_timer_bias;
	/* Stable timer control KVM register */
	u32 stable_timer_ctl;
	/* Dynamic nanosecond bias (multiple of stable_timer_period) to avoid overflow */
	s64 stable_timer_dyn_bias;
	/* Resume time */
	ktime_t stable_timer_resume;
	/* Save ktime */
	ktime_t stable_ktime_saved;

	u64 core_ext_ioisr[4];

	/* Last CPU the VCPU state was loaded on */
	int last_sched_cpu;
	/* Last CPU the VCPU actually executed guest code on */
	int last_exec_cpu;

	/* WAIT executed */
	int wait;

	u8 fpu_enabled;
	u8 lsx_enabled;
};

static inline void _kvm_atomic_set_guest_reg(unsigned long *reg,
						unsigned long val)
{
	unsigned long temp;
	do {
		__asm__ __volatile__(
		"	" __LL "%0, %1				\n"
		"	or	%0, %2				\n"
		"	" __SC	"%0, %1				\n"
		: "=&r" (temp), "+m" (*reg)
		: "r" (val));
	} while (unlikely(!temp));
}

static inline void _kvm_atomic_clear_guest_reg(unsigned long *reg,
						  unsigned long val)
{
	unsigned long temp;
	do {
		__asm__ __volatile__(
		"	" __LL "%0, %1				\n"
		"	and	%0, %2				\n"
		"	" __SC	"%0, %1				\n"
		: "=&r" (temp), "+m" (*reg)
		: "r" (~val));
	} while (unlikely(!temp));
}

static inline void _kvm_atomic_change_guest_reg(unsigned long *reg,
						   unsigned long change,
						   unsigned long val)
{
	unsigned long temp;
	do {
		__asm__ __volatile__(
		"	" __LL "%0, %1				\n"
		"	and	%0, %2				\n"
		"	or	%0, %3				\n"
		"	" __SC	"%0, %1				\n"
		: "=&r" (temp), "+m" (*reg)
		: "r" (~change), "r" (val & change));
	} while (unlikely(!temp));
}

/* Guest register types, used in accessor build below */
#define __KVMT32	u32
#define __KVMTl	unsigned long

/*
 * __BUILD_KVM_$ops_SAVED(): kvm_$op_sw_csr_$reg()
 * These operate on the saved guest C0 state in RAM.
 */

/* Generate saved context simple accessors */
#define __BUILD_KVM_RW_SAVED(name, type, _reg)				\
static inline __KVMT##type kvm_read_sw_gcsr_##name(struct loongarch_csrs *csr) \
{									\
	return csr->csrs[(_reg)];					\
}									\
static inline void kvm_write_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					   __KVMT##type val)		\
{									\
	csr->csrs[(_reg)] = val;					\
}

/* Generate saved context bitwise modifiers */
#define __BUILD_KVM_SET_SAVED(name, type, _reg)				\
static inline void kvm_set_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					 __KVMT##type val)		\
{									\
	csr->csrs[(_reg)] |= val;					\
}									\
static inline void kvm_clear_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					   __KVMT##type val)		\
{									\
	csr->csrs[(_reg)] &= ~val;					\
}									\
static inline void kvm_change_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					    __KVMT##type mask,		\
					    __KVMT##type val)		\
{									\
	unsigned long _mask = mask;					\
	csr->csrs[(_reg)] &= ~_mask;					\
	csr->csrs[(_reg)] |= val & _mask;				\
}

/* Generate saved context atomic bitwise modifiers */
#define __BUILD_KVM_ATOMIC_SAVED(name, type, _reg)			\
static inline void kvm_set_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					 __KVMT##type val)		\
{									\
	_kvm_atomic_set_guest_reg(&csr->csrs[(_reg)], val);		\
}									\
static inline void kvm_clear_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					   __KVMT##type val)		\
{									\
	_kvm_atomic_clear_guest_reg(&csr->csrs[(_reg)], val);		\
}									\
static inline void kvm_change_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					    __KVMT##type mask,		\
					    __KVMT##type val)		\
{									\
	_kvm_atomic_change_guest_reg(&csr->csrs[(_reg)], mask, val);	\
}

/* Uncached window access based on CSR_UCWIN_BASE */
#define __BUILD_KVM_RW_SAVED_UCA(name, type, _reg)			\
static inline __KVMT##type kvm_read_sw_gcsr_##name(struct loongarch_csrs *csr) \
{									\
	return csr->csr_uca[(_reg) - CSR_UCWIN_BASE];			\
}									\
static inline void kvm_write_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					   __KVMT##type val)		\
{									\
	csr->csr_uca[(_reg) - CSR_UCWIN_BASE] = val;			\
}

/* Generate saved context bitwise modifiers */
#define __BUILD_KVM_SET_SAVED_UCA(name, type, _reg)			\
static inline void kvm_set_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					 __KVMT##type val)		\
{									\
	csr->csr_uca[(_reg) - CSR_UCWIN_BASE] |= val;			\
}									\
static inline void kvm_clear_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					   __KVMT##type val)		\
{									\
	csr->csr_uca[(_reg) - CSR_UCWIN_BASE] &= ~val;			\
}									\
static inline void kvm_change_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					    __KVMT##type mask,		\
					    __KVMT##type val)		\
{									\
	unsigned long _mask = mask;					\
	csr->csr_uca[(_reg) - CSR_UCWIN_BASE] &= ~_mask;		\
	csr->csr_uca[(_reg) - CSR_UCWIN_BASE] |= val & _mask;		\
}

/* Generate saved context atomic bitwise modifiers */
#define __BUILD_KVM_ATOMIC_SAVED_UCA(name, type, _reg)			\
static inline void kvm_set_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					 __KVMT##type val)		\
{									\
	_kvm_atomic_set_guest_reg(&csr->csr_uca[(_reg) -		\
						CSR_UCWIN_BASE], val);	\
}									\
static inline void kvm_clear_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					   __KVMT##type val)		\
{									\
	_kvm_atomic_clear_guest_reg(&csr->csr_uca[(_reg) -		\
						CSR_UCWIN_BASE], val);	\
}									\
static inline void kvm_change_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					    __KVMT##type mask,		\
					    __KVMT##type val)		\
{									\
	_kvm_atomic_change_guest_reg(&csr->csr_uca[(_reg) -		\
					CSR_UCWIN_BASE], mask, val);	\
}

/* Generate saved context simple accessors */
#define __BUILD_KVM_RW_SAVED_DM(name, type, _reg)			\
static inline __KVMT##type kvm_read_sw_gcsr_##name(struct loongarch_csrs *csr) \
{									\
	return csr->csr_dmwin[(_reg) - CSR_DMWIN_BASE];			\
}									\
static inline void kvm_write_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					   __KVMT##type val)		\
{									\
	csr->csr_dmwin[(_reg) - CSR_DMWIN_BASE] = val;			\
}

/* Generate saved context bitwise modifiers */
#define __BUILD_KVM_SET_SAVED_DM(name, type, _reg)			\
static inline void kvm_set_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					 __KVMT##type val)		\
{									\
	csr->csr_dmwin[(_reg) - CSR_DMWIN_BASE] |= val;			\
}									\
static inline void kvm_clear_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					   __KVMT##type val)		\
{									\
	csr->csr_dmwin[(_reg) - CSR_DMWIN_BASE] &= ~val;		\
}									\
static inline void kvm_change_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					    __KVMT##type mask,		\
					    __KVMT##type val)		\
{									\
	unsigned long _mask = mask;					\
	csr->csr_dmwin[(_reg) - CSR_DMWIN_BASE] &= ~_mask;		\
	csr->csr_dmwin[(_reg) - CSR_DMWIN_BASE] |= val & _mask;		\
}

/* debug access based on 0x100 */
#define __BUILD_KVM_RW_SAVED_DBG(name, type, _reg)			\
static inline __KVMT##type kvm_read_sw_gcsr_##name(struct loongarch_csrs *csr) \
{									\
	return csr->csr_dbg[(_reg) - CSR_DEBUG_BASE];			\
}									\
static inline void kvm_write_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					   __KVMT##type val)		\
{									\
	csr->csr_dbg[(_reg) - CSR_DEBUG_BASE] = val;			\
}

/* Generate saved context bitwise modifiers */
#define __BUILD_KVM_SET_SAVED_DBG(name, type, _reg)			\
static inline void kvm_set_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					 __KVMT##type val)		\
{									\
	csr->csr_debug[(_reg) - CSR_DEBUG_BASE] |= val;			\
}									\
static inline void kvm_clear_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					   __KVMT##type val)		\
{									\
	csr->csr_debug[(_reg) - CSR_DEBUG_BASE] &= ~val;		\
}									\
static inline void kvm_change_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					    __KVMT##type mask,		\
					    __KVMT##type val)		\
{									\
	unsigned long _mask = mask;					\
	csr->csr_debug[(_reg) - CSR_DEBUG_BASE] &= ~_mask;		\
	csr->csr_debug[(_reg) - CSR_DEBUG_BASE] |= val & _mask;		\
}

/* Generate saved context atomic bitwise modifiers */
#define __BUILD_KVM_ATOMIC_SAVED_DBG(name, type, _reg)			\
static inline void kvm_set_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					 __KVMT##type val)		\
{									\
	_kvm_atomic_set_guest_reg(&csr->csr_dbg[(_reg) -		\
						CSR_DEBUG_BASE], val);	\
}									\
static inline void kvm_clear_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					   __KVMT##type val)		\
{									\
	_kvm_atomic_clear_guest_reg(&csr->csr_dbg[(_reg) -		\
						CSR_DEBUG_BASE], val);	\
}									\
static inline void kvm_change_sw_gcsr_##name(struct loongarch_csrs *csr,	\
					    __KVMT##type mask,		\
					    __KVMT##type val)		\
{									\
	_kvm_atomic_change_guest_reg(&csr->csr_dbg[(_reg) -		\
					CSR_DEBUG_BASE], mask, val);	\
}

/*
 * __BUILD_KVM_$ops_VZ(): kvm_$op_vz_csr_$reg()
 * These operate on the VZ guest C0 context in hardware.
 */

/* Generate VZ guest context simple accessors */
#define __BUILD_KVM_RW_VZ(name, type, _reg)				\
static inline __KVMT##type kvm_read_vz_gcsr_##name(struct loongarch_csrs *csr) \
{									\
	return read_gcsr_##name();					\
}									\
static inline void kvm_write_vz_gcsr_##name(struct loongarch_csrs *csr,	\
					   __KVMT##type val)		\
{									\
	write_gcsr_##name(val);						\
}

/* Generate VZ guest context bitwise modifiers */
#define __BUILD_KVM_SET_VZ(name, type, _reg)				\
static inline void kvm_set_vz_gcsr_##name(struct loongarch_csrs *csr,	\
					 __KVMT##type val)		\
{									\
	set_gcsr_##name(val);						\
}									\
static inline void kvm_clear_vz_gcsr_##name(struct loongarch_csrs *csr,	\
					   __KVMT##type val)		\
{									\
	clear_gcsr_##name(val);						\
}									\
static inline void kvm_change_vz_gcsr_##name(struct loongarch_csrs *csr,	\
					    __KVMT##type mask,		\
					    __KVMT##type val)		\
{									\
	change_gcsr_##name(mask, val);					\
}

/* Generate VZ guest context save/restore to/from saved context */
#define __BUILD_KVM_SAVE_VZ(name, _reg)					\
static inline void kvm_restore_gcsr_##name(struct loongarch_csrs *csr)	\
{									\
	write_gcsr_##name(csr->csrs[(_reg)]);				\
}									\
static inline void kvm_save_gcsr_##name(struct loongarch_csrs *csr)	\
{									\
	csr->csrs[(_reg)] = read_gcsr_##name();				\
}

/* Direct mapped window CSRS access */
#define __BUILD_KVM_SAVE_VZ_DM(name, _reg)				\
static inline void kvm_restore_gcsr_##name(struct loongarch_csrs *csr)	\
{									\
	write_gcsr_##name(csr->csr_dmwin[(_reg) - CSR_DMWIN_BASE]);	\
}									\
static inline void kvm_save_gcsr_##name(struct loongarch_csrs *csr)	\
{									\
	csr->csr_dmwin[(_reg) - CSR_DMWIN_BASE] = read_gcsr_##name();	\
}

/*
 * __BUILD_KVM_$ops_WRAP(): kvm_$op_$name1() -> kvm_$op_$name2()
 * These wrap a set of operations to provide them with a different name.
 */

/* Generate simple accessor wrapper */
#define __BUILD_KVM_RW_WRAP(name1, name2, type)				\
static inline __KVMT##type kvm_read_##name1(struct loongarch_csrs *csr)	\
{									\
	return kvm_read_##name2(csr);					\
}									\
static inline void kvm_write_##name1(struct loongarch_csrs *csr,		\
				     __KVMT##type val)			\
{									\
	kvm_write_##name2(csr, val);					\
}

/* Generate bitwise modifier wrapper */
#define __BUILD_KVM_SET_WRAP(name1, name2, type)			\
static inline void kvm_set_##name1(struct loongarch_csrs *csr,		\
				   __KVMT##type val)			\
{									\
	kvm_set_##name2(csr, val);					\
}									\
static inline void kvm_clear_##name1(struct loongarch_csrs *csr,		\
				     __KVMT##type val)			\
{									\
	kvm_clear_##name2(csr, val);					\
}									\
static inline void kvm_change_##name1(struct loongarch_csrs *csr,	\
				      __KVMT##type mask,		\
				      __KVMT##type val)			\
{									\
	kvm_change_##name2(csr, mask, val);				\
}

/*
 * __BUILD_KVM_$ops_SW(): kvm_$op_csr_guest_$reg() -> kvm_$op_sw_csr_$reg()
 * These generate accessors operating on the saved context in RAM, and wrap them
 * with the common guest C0 accessors (for use by common emulation code).
 */

#define __BUILD_KVM_RW_SW(name, type, _reg)				\
	__BUILD_KVM_RW_SAVED(name, type, _reg)				\
	__BUILD_KVM_RW_WRAP(csr_guest_##name, sw_gcsr_##name, type)

#define __BUILD_KVM_SET_SW(name, type, _reg)				\
	__BUILD_KVM_SET_SAVED(name, type, _reg)				\
	__BUILD_KVM_SET_WRAP(csr_guest_##name, sw_gcsr_##name, type)

#define __BUILD_KVM_ATOMIC_SW(name, type, _reg)				\
	__BUILD_KVM_ATOMIC_SAVED(name, type, _reg)			\
	__BUILD_KVM_SET_WRAP(csr_guest_##name, sw_gcsr_##name, type)

/* Uncached window csrs access */
#define __BUILD_KVM_RW_SW_UCA(name, type, _reg)				\
	__BUILD_KVM_RW_SAVED_UCA(name, type, _reg)			\
	__BUILD_KVM_RW_WRAP(csr_guest_##name, sw_gcsr_##name, type)

#define __BUILD_KVM_SET_SW_UCA(name, type, _reg)			\
	__BUILD_KVM_SET_SAVED_UCA(name, type, _reg)			\
	__BUILD_KVM_SET_WRAP(csr_guest_##name, sw_gcsr_##name, type)

#define __BUILD_KVM_ATOMIC_SW_UCA(name, type, _reg)			\
	__BUILD_KVM_ATOMIC_SAVED_UCA(name, type, _reg)			\
	__BUILD_KVM_SET_WRAP(csr_guest_##name, sw_gcsr_##name, type)

/* Debug related csrs access */
#define __BUILD_KVM_RW_SW_DBG(name, type, _reg)				\
	__BUILD_KVM_RW_SAVED_DBG(name, type, _reg)			\
	__BUILD_KVM_RW_WRAP(csr_guest_##name, sw_gcsr_##name, type)

#define __BUILD_KVM_SET_SW_DBG(name, type, _reg)			\
	__BUILD_KVM_SET_SAVED_DBG(name, type, _reg)			\
	__BUILD_KVM_SET_WRAP(csr_guest_##name, sw_gcsr_##name, type)

#define __BUILD_KVM_ATOMIC_SW_DBG(name, type, _reg)			\
	__BUILD_KVM_ATOMIC_SAVED_DBG(name, type, _reg)			\
	__BUILD_KVM_SET_WRAP(csr_guest_##name, sw_gcsr_##name, type)

/*
 * VZ (hardware assisted virtualisation)
 * These macros use the active guest state in VZ mode (hardware registers),
 */

/*
 * __BUILD_KVM_$ops_HW(): kvm_$op_csr_guest_$reg() -> kvm_$op_vz_csr_$reg()
 * These generate accessors operating on the VZ guest context in hardware, and
 * wrap them with the common guest C0 accessors (for use by common emulation
 * code).
 *
 * Accessors operating on the saved context in RAM are also generated to allow
 * convenient explicit saving and restoring of the state.
 */

#define __BUILD_KVM_RW_HW(name, type, _reg)				\
	__BUILD_KVM_RW_SAVED(name, type, _reg)				\
	__BUILD_KVM_RW_VZ(name, type, _reg)				\
	__BUILD_KVM_RW_WRAP(csr_guest_##name, vz_gcsr_##name, type)	\
	__BUILD_KVM_SAVE_VZ(name, _reg)

#define __BUILD_KVM_SET_HW(name, type, _reg)				\
	__BUILD_KVM_SET_SAVED(name, type, _reg)				\
	__BUILD_KVM_SET_VZ(name, type, _reg)				\
	__BUILD_KVM_SET_WRAP(csr_guest_##name, vz_gcsr_##name, type)

/* Direct mapped window CSRs access */
#define __BUILD_KVM_RW_HW_DM(name, type, _reg)				\
	__BUILD_KVM_RW_SAVED_DM(name, type, _reg)			\
	__BUILD_KVM_RW_VZ(name, type, _reg)				\
	__BUILD_KVM_RW_WRAP(csr_guest_##name, vz_gcsr_##name, type)	\
	__BUILD_KVM_SAVE_VZ_DM(name, _reg)

#define __BUILD_KVM_SET_HW_DM(name, type, _reg)				\
	__BUILD_KVM_SET_SAVED_DM(name, type, _reg)			\
	__BUILD_KVM_SET_VZ(name, type, _reg)				\

/*
 * We can't do atomic modifications of COP0 state if hardware can modify it.
 * Races must be handled explicitly.
 */
#define __BUILD_KVM_ATOMIC_HW		__BUILD_KVM_SET_HW
#define __BUILD_KVM_ATOMIC_HW_DM	__BUILD_KVM_SET_HW_DM

/*
 * Define accessors for CP0 registers that are accessible to the guest. These
 * are primarily used by common emulation code, which may need to access the
 * registers differently depending on the implementation.
 *
 *    fns_hw/sw    name     type    reg num         select
 */
__BUILD_KVM_RW_HW(crmd,           l, KVM_LOONGARCH_CSR_CRMD)
__BUILD_KVM_RW_HW(prmd,           l, KVM_LOONGARCH_CSR_PRMD)
__BUILD_KVM_RW_HW(coprocessor,    l, KVM_LOONGARCH_CSR_EUEN)
__BUILD_KVM_RW_HW(misc,           l, KVM_LOONGARCH_CSR_MISC)
__BUILD_KVM_RW_HW(ecfg,           l, KVM_LOONGARCH_CSR_ECFG)
__BUILD_KVM_RW_HW(estat,          l, KVM_LOONGARCH_CSR_ESTAT)
__BUILD_KVM_RW_HW(epc,            l, KVM_LOONGARCH_CSR_EPC)
__BUILD_KVM_RW_HW(badv,           l, KVM_LOONGARCH_CSR_BADV)
__BUILD_KVM_RW_HW(badi,           l, KVM_LOONGARCH_CSR_BADI)
__BUILD_KVM_RW_HW(ebase,          l, KVM_LOONGARCH_CSR_EBASE)
__BUILD_KVM_RW_HW(tlbidx,         l, KVM_LOONGARCH_CSR_TLBIDX)
__BUILD_KVM_RW_HW(tlbhi,          l, KVM_LOONGARCH_CSR_TLBHI)
__BUILD_KVM_RW_HW(tlblo0,         l, KVM_LOONGARCH_CSR_TLBLO0)
__BUILD_KVM_RW_HW(tlblo1,         l, KVM_LOONGARCH_CSR_TLBLO1)
__BUILD_KVM_RW_HW(asid,           l, KVM_LOONGARCH_CSR_ASID)
__BUILD_KVM_RW_HW(pgdl,           l, KVM_LOONGARCH_CSR_PGDL)
__BUILD_KVM_RW_HW(pgdh,           l, KVM_LOONGARCH_CSR_PGDH)
__BUILD_KVM_RW_HW(pwctl0,         l, KVM_LOONGARCH_CSR_PWCTL0)
__BUILD_KVM_RW_HW(pwctl1,         l, KVM_LOONGARCH_CSR_PWCTL1)
__BUILD_KVM_RW_HW(stlbpgsize,     l, KVM_LOONGARCH_CSR_STLBPS)
__BUILD_KVM_RW_HW(rvacfg,         l, KVM_LOONGARCH_CSR_RVACFG)
__BUILD_KVM_RW_HW(cpunum,         l, KVM_LOONGARCH_CSR_CPUNUM)
__BUILD_KVM_RW_HW(prcfg1,         l, KVM_LOONGARCH_CSR_PRCFG1)
__BUILD_KVM_RW_HW(prcfg2,         l, KVM_LOONGARCH_CSR_PRCFG2)
__BUILD_KVM_RW_HW(prcfg3,         l, KVM_LOONGARCH_CSR_PRCFG3)
__BUILD_KVM_RW_HW(kscratch0,      l, KVM_LOONGARCH_CSR_KSCRATCH0)
__BUILD_KVM_RW_HW(kscratch1,      l, KVM_LOONGARCH_CSR_KSCRATCH1)
__BUILD_KVM_RW_HW(kscratch2,      l, KVM_LOONGARCH_CSR_KSCRATCH2)
__BUILD_KVM_RW_HW(kscratch3,      l, KVM_LOONGARCH_CSR_KSCRATCH3)
__BUILD_KVM_RW_HW(kscratch4,      l, KVM_LOONGARCH_CSR_KSCRATCH4)
__BUILD_KVM_RW_HW(kscratch5,      l, KVM_LOONGARCH_CSR_KSCRATCH5)
__BUILD_KVM_RW_HW(kscratch6,      l, KVM_LOONGARCH_CSR_KSCRATCH6)
__BUILD_KVM_RW_HW(kscratch7,      l, KVM_LOONGARCH_CSR_KSCRATCH7)
__BUILD_KVM_RW_HW(timerid,        l, KVM_LOONGARCH_CSR_TIMERID)
__BUILD_KVM_RW_HW(timercfg,       l, KVM_LOONGARCH_CSR_TIMERCFG)
__BUILD_KVM_RW_HW(timertick,      l, KVM_LOONGARCH_CSR_TIMERTICK)
__BUILD_KVM_RW_HW(timeroffset,    l, KVM_LOONGARCH_CSR_TIMEROFFSET)
__BUILD_KVM_RW_HW(llbctl,         l, KVM_LOONGARCH_CSR_LLBCTL)
__BUILD_KVM_RW_HW(tlbrebase,      l, KVM_LOONGARCH_CSR_TLBREBASE)
__BUILD_KVM_RW_HW(tlbrbadv,       l, KVM_LOONGARCH_CSR_TLBRBADV)
__BUILD_KVM_RW_HW(tlbrepc,        l, KVM_LOONGARCH_CSR_TLBREPC)
__BUILD_KVM_RW_HW(tlbrsave,       l, KVM_LOONGARCH_CSR_TLBRSAVE)
__BUILD_KVM_RW_HW(tlbrelo0,       l, KVM_LOONGARCH_CSR_TLBRELO0)
__BUILD_KVM_RW_HW(tlbrelo1,       l, KVM_LOONGARCH_CSR_TLBRELO1)
__BUILD_KVM_RW_HW(tlbrehi,        l, KVM_LOONGARCH_CSR_TLBREHI)
__BUILD_KVM_RW_HW(tlbrprmd,       l, KVM_LOONGARCH_CSR_TLBRPRMD)

__BUILD_KVM_RW_SW(gtlbc,          l, KVM_LOONGARCH_CSR_GTLBC)
__BUILD_KVM_RW_SW(trgp,           l, KVM_LOONGARCH_CSR_TRGP)
__BUILD_KVM_RW_SW(gstat,          l, KVM_LOONGARCH_CSR_GSTAT)
__BUILD_KVM_RW_SW(gcfg,           l, KVM_LOONGARCH_CSR_GCFG)
__BUILD_KVM_RW_SW(gintc,          l, KVM_LOONGARCH_CSR_GINTC)
__BUILD_KVM_RW_SW(impctl1,        l, KVM_LOONGARCH_CSR_IMPCTL1)
__BUILD_KVM_RW_SW(impctl2,        l, KVM_LOONGARCH_CSR_IMPCTL2)
__BUILD_KVM_RW_SW(gnmi,           l, KVM_LOONGARCH_CSR_GNMI)
__BUILD_KVM_RW_SW(errctl,         l, KVM_LOONGARCH_CSR_ERRCTL)
__BUILD_KVM_RW_SW(errinfo1,       l, KVM_LOONGARCH_CSR_ERRINFO1)
__BUILD_KVM_RW_SW(errinfo2,       l, KVM_LOONGARCH_CSR_ERRINFO2)
__BUILD_KVM_RW_SW(errebase,       l, KVM_LOONGARCH_CSR_ERREBASE)
__BUILD_KVM_RW_SW(errepc,         l, KVM_LOONGARCH_CSR_ERREPC)
__BUILD_KVM_RW_SW(errsave,        l, KVM_LOONGARCH_CSR_ERRSAVE)
__BUILD_KVM_RW_SW(ctag,           l, KVM_LOONGARCH_CSR_CTAG)
__BUILD_KVM_RW_SW(mcsr0,          l, KVM_LOONGARCH_CSR_MCSR0)
__BUILD_KVM_RW_SW(mcsr1,          l, KVM_LOONGARCH_CSR_MCSR1)
__BUILD_KVM_RW_SW(mcsr2,          l, KVM_LOONGARCH_CSR_MCSR2)
__BUILD_KVM_RW_SW(mcsr3,          l, KVM_LOONGARCH_CSR_MCSR3)
__BUILD_KVM_RW_SW(mcsr8,          l, KVM_LOONGARCH_CSR_MCSR8)
__BUILD_KVM_RW_SW(mcsr9,          l, KVM_LOONGARCH_CSR_MCSR9)
__BUILD_KVM_RW_SW(mcsr10,         l, KVM_LOONGARCH_CSR_MCSR10)
__BUILD_KVM_RW_SW(mcsr24,         l, KVM_LOONGARCH_CSR_MCSR24)

/* uncached accelerate windows */
__BUILD_KVM_RW_SW_UCA(ucawin,     l, KVM_LOONGARCH_CSR_UCAWIN)
__BUILD_KVM_RW_SW_UCA(ucawin0lo,  l, KVM_LOONGARCH_CSR_UCAWIN0_LO)
__BUILD_KVM_RW_SW_UCA(ucawin0hi,  l, KVM_LOONGARCH_CSR_UCAWIN0_HI)
__BUILD_KVM_RW_SW_UCA(ucawin1lo,  l, KVM_LOONGARCH_CSR_UCAWIN1_LO)
__BUILD_KVM_RW_SW_UCA(ucawin1hi,  l, KVM_LOONGARCH_CSR_UCAWIN1_HI)
__BUILD_KVM_RW_SW_UCA(ucawin2lo,  l, KVM_LOONGARCH_CSR_UCAWIN2_LO)
__BUILD_KVM_RW_SW_UCA(ucawin2hi,  l, KVM_LOONGARCH_CSR_UCAWIN2_HI)
__BUILD_KVM_RW_SW_UCA(ucawin3lo,  l, KVM_LOONGARCH_CSR_UCAWIN3_LO)
__BUILD_KVM_RW_SW_UCA(ucawin3hi,  l, KVM_LOONGARCH_CSR_UCAWIN3_HI)
__BUILD_KVM_RW_SW_DBG(debug,      l, KVM_LOONGARCH_CSR_DEBUG)
__BUILD_KVM_RW_SW_DBG(desave,     l, KVM_LOONGARCH_CSR_DESAVE)
__BUILD_KVM_RW_SW_DBG(depc,       l, KVM_LOONGARCH_CSR_DEPC)

/* direct map windows */
__BUILD_KVM_RW_HW_DM(directwin0,  l, KVM_LOONGARCH_CSR_DMWIN0)
__BUILD_KVM_RW_HW_DM(directwin1,  l, KVM_LOONGARCH_CSR_DMWIN1)
__BUILD_KVM_RW_HW_DM(directwin2,  l, KVM_LOONGARCH_CSR_DMWIN2)
__BUILD_KVM_RW_HW_DM(directwin3,  l, KVM_LOONGARCH_CSR_DMWIN3)

/* Bitwise operations (on HW state) */
/* Cause can be modified asynchronously from hardirq hrtimer callback */
//__BUILD_KVM_ATOMIC_HW(cause,      32, LOONGARCH_CSR_CAUSE)

/* Bitwise operations (on saved state) */
__BUILD_KVM_SET_SW(impctl1,        l, KVM_LOONGARCH_CSR_IMPCTL1)
__BUILD_KVM_SET_SW(impctl2,        l, KVM_LOONGARCH_CSR_IMPCTL2)
__BUILD_KVM_SET_SW(mcsr0,          l, KVM_LOONGARCH_CSR_MCSR0)
__BUILD_KVM_SET_SW(mcsr1,          l, KVM_LOONGARCH_CSR_MCSR1)
__BUILD_KVM_SET_SW(mcsr2,          l, KVM_LOONGARCH_CSR_MCSR2)
__BUILD_KVM_SET_SW(mcsr3,          l, KVM_LOONGARCH_CSR_MCSR3)
__BUILD_KVM_SET_SW(mcsr8,          l, KVM_LOONGARCH_CSR_MCSR8)
__BUILD_KVM_SET_SW(mcsr9,          l, KVM_LOONGARCH_CSR_MCSR9)
__BUILD_KVM_SET_SW(mcsr10,         l, KVM_LOONGARCH_CSR_MCSR10)
__BUILD_KVM_SET_SW(mcsr24,         l, KVM_LOONGARCH_CSR_MCSR24)
__BUILD_KVM_SET_SW_UCA(ucawin,     l, KVM_LOONGARCH_CSR_UCAWIN)
__BUILD_KVM_SET_SW_UCA(ucawin0lo,  l, KVM_LOONGARCH_CSR_UCAWIN0_LO)
__BUILD_KVM_SET_SW_UCA(ucawin0hi,  l, KVM_LOONGARCH_CSR_UCAWIN0_HI)
__BUILD_KVM_SET_SW_UCA(ucawin1lo,  l, KVM_LOONGARCH_CSR_UCAWIN1_LO)
__BUILD_KVM_SET_SW_UCA(ucawin1hi,  l, KVM_LOONGARCH_CSR_UCAWIN1_HI)
__BUILD_KVM_SET_SW_UCA(ucawin2lo,  l, KVM_LOONGARCH_CSR_UCAWIN2_LO)
__BUILD_KVM_SET_SW_UCA(ucawin2hi,  l, KVM_LOONGARCH_CSR_UCAWIN2_HI)
__BUILD_KVM_SET_SW_UCA(ucawin3lo,  l, KVM_LOONGARCH_CSR_UCAWIN3_LO)
__BUILD_KVM_SET_SW_UCA(ucawin3hi,  l, KVM_LOONGARCH_CSR_UCAWIN3_HI)


/* Helpers */

static inline bool kvm_loongarch_guest_can_have_fpu(struct kvm_vcpu_arch *arch)
{
	return (!__builtin_constant_p(raw_cpu_has_fpu) || raw_cpu_has_fpu) &&
		arch->fpu_enabled;
}

static inline bool kvm_loongarch_guest_has_fpu(struct kvm_vcpu_arch *arch)
{
	return kvm_loongarch_guest_can_have_fpu(arch);
}

static inline bool kvm_loongarch_guest_can_have_lsx(struct kvm_vcpu_arch *arch)
{
	return (!__builtin_constant_p(cpu_has_lsx) || cpu_has_lsx) &&
		arch->lsx_enabled;
}

static inline bool kvm_loongarch_guest_has_lsx(struct kvm_vcpu_arch *arch)
{
	return kvm_loongarch_guest_can_have_lsx(arch);
}

bool kvm_loongarch_guest_has_lasx(struct kvm_vcpu *vcpu);

struct kvm_loongarch_callbacks {
	int (*handle_tlb_ld_miss)(struct kvm_vcpu *vcpu);
	int (*handle_tlb_st_miss)(struct kvm_vcpu *vcpu);
	int (*handle_tlbifetch)(struct kvm_vcpu *vcpu);
	int (*handle_tlb_mod)(struct kvm_vcpu *vcpu);
	int (*handle_tlbri)(struct kvm_vcpu *vcpu);
	int (*handle_tlbxi)(struct kvm_vcpu *vcpu);
	int (*handle_tlbprev)(struct kvm_vcpu *vcpu);
	int (*handle_addr)(struct kvm_vcpu *vcpu);
	int (*handle_unalign)(struct kvm_vcpu *vcpu);
	int (*handle_bound)(struct kvm_vcpu *vcpu);
	int (*handle_syscall)(struct kvm_vcpu *vcpu);
	int (*handle_break)(struct kvm_vcpu *vcpu);
	int (*handle_res_inst)(struct kvm_vcpu *vcpu);
	int (*handle_prev_inst)(struct kvm_vcpu *vcpu);
	int (*handle_fp_disabled)(struct kvm_vcpu *vcpu);
	int (*handle_lsx_disabled)(struct kvm_vcpu *vcpu);
	int (*handle_lasx_disabled)(struct kvm_vcpu *vcpu);
	int (*handle_fpe)(struct kvm_vcpu *vcpu);
	int (*handle_watch)(struct kvm_vcpu *vcpu);
	int (*handle_psi)(struct kvm_vcpu *vcpu);
	int (*handle_hypcall)(struct kvm_vcpu *vcpu);
	int (*handle_field_change)(struct kvm_vcpu *vcpu);
	int (*hardware_enable)(void);
	void (*hardware_disable)(void);
	int (*check_extension)(struct kvm *kvm, long ext);
	int (*vcpu_init)(struct kvm_vcpu *vcpu);
	void (*vcpu_uninit)(struct kvm_vcpu *vcpu);
	int (*vcpu_setup)(struct kvm_vcpu *vcpu);
	void (*flush_shadow_all)(struct kvm *kvm);
	/*
	 * Must take care of flushing any cached GPA PTEs (e.g. guest entries in
	 * VZ root TLB, or T&E GVA page tables and corresponding root TLB
	 * mappings).
	 */
	void (*flush_shadow_memslot)(struct kvm *kvm,
				     const struct kvm_memory_slot *slot);
	gpa_t (*gva_to_gpa)(gva_t gva);
	void (*queue_timer_int)(struct kvm_vcpu *vcpu);
	void (*dequeue_timer_int)(struct kvm_vcpu *vcpu);
	void (*queue_io_int)(struct kvm_vcpu *vcpu,
			     struct kvm_loongarch_interrupt *irq);
	void (*dequeue_io_int)(struct kvm_vcpu *vcpu,
			       struct kvm_loongarch_interrupt *irq);
	int (*irq_deliver)(struct kvm_vcpu *vcpu, unsigned int priority,
			   u64 exst);
	int (*irq_clear)(struct kvm_vcpu *vcpu, unsigned int priority,
			 u64 exst);
	unsigned long (*num_regs)(struct kvm_vcpu *vcpu);
	int (*copy_reg_indices)(struct kvm_vcpu *vcpu, u64 __user *indices);
	int (*get_one_reg)(struct kvm_vcpu *vcpu,
			   const struct kvm_one_reg *reg, s64 *v);
	int (*set_one_reg)(struct kvm_vcpu *vcpu,
			   const struct kvm_one_reg *reg, s64 v);
	int (*vcpu_load)(struct kvm_vcpu *vcpu, int cpu);
	int (*vcpu_put)(struct kvm_vcpu *vcpu, int cpu);
	int (*vcpu_run)(struct kvm_run *run, struct kvm_vcpu *vcpu);
	void (*vcpu_reenter)(struct kvm_run *run, struct kvm_vcpu *vcpu);
};
extern struct kvm_loongarch_callbacks *kvm_loongarch_callbacks;
int kvm_loongarch_emulation_init(struct kvm_loongarch_callbacks **install_callbacks);

struct kvm_timer_callbacks {
	void (*restore_timer)(struct kvm_vcpu *vcpu);
	void (*acquire_htimer)(struct kvm_vcpu *vcpu);
	void (*save_timer)(struct kvm_vcpu *vcpu);
	void (*lose_htimer)(struct kvm_vcpu *vcpu);
	enum hrtimer_restart (*count_timeout)(struct kvm_vcpu *vcpu);
	void (*write_count)(struct kvm_vcpu *vcpu, u32 count);
	void (*write_stable_timer)(struct kvm_vcpu *vcpu, u64 stable_timer);
};
extern struct kvm_timer_callbacks *kvm_timer_callbacks;
int kvm_loongarch_timer_init(struct kvm_timer_callbacks **install_callbacks);

/* Debug: dump vcpu state */
int kvm_arch_vcpu_dump_regs(struct kvm_vcpu *vcpu);

extern int kvm_loongarch_handle_exit(struct kvm_run *run, struct kvm_vcpu *vcpu);

/* Building of entry/exception code */
int kvm_loongarch_scratch_setup(void);
void *kvm_loongarch_build_vcpu_run(void *addr);
void *kvm_loongarch_build_tlb_refill_exception(void *addr, void *handler);
void *kvm_loongarch_build_exception(void *addr, void *handler);
void *kvm_loongarch_build_exit(void *addr);

/* TLB handlings */
int kvm_loongarch_handle_lvz_root_tlb_fault(unsigned long badvaddr,
				      struct kvm_vcpu *vcpu, bool write_fault);
int kvm_lvz_host_tlb_inv(struct kvm_vcpu *vcpu, unsigned long entryhi);
int kvm_lvz_host_tlb_update(struct kvm_vcpu *vcpu, unsigned long entryhi);
int kvm_lvz_guest_tlb_lookup(struct kvm_vcpu *vcpu, unsigned long gva,
			    unsigned long *gpa);
void kvm_lvz_local_flush_roottlb_all_guests(void);
void kvm_lvz_local_flush_guesttlb_all(void);
void kvm_lvz_save_guesttlb(struct kvm_loongarch_tlb *buf, unsigned int index,
			  unsigned int count);
void kvm_lvz_load_guesttlb(const struct kvm_loongarch_tlb *buf, unsigned int index,
			  unsigned int count);
void kvm_loongarch_clear_guest_mtlb(void);
void kvm_loongarch_clear_guest_stlb(void);

/* MMU handling */

bool kvm_loongarch_flush_gpa_pt(struct kvm *kvm, gfn_t start_gfn, gfn_t end_gfn);
int kvm_loongarch_mkclean_gpa_pt(struct kvm *kvm, gfn_t start_gfn, gfn_t end_gfn);
pgd_t *kvm_pgd_alloc(void);
void kvm_mmu_free_memory_caches(struct kvm_vcpu *vcpu);

enum kvm_loongarch_fault_result {
	KVM_LOONGARCH_MAPPED = 0,
	KVM_LOONGARCH_GVA,
	KVM_LOONGARCH_GPA,
	KVM_LOONGARCH_TLB,
	KVM_LOONGARCH_TLBINV,
	KVM_LOONGARCH_TLBMOD,
};
enum kvm_loongarch_fault_result kvm_trap_emul_gva_fault(struct kvm_vcpu *vcpu,
						   unsigned long gva,
						   bool write);
#define KVM_ARCH_WANT_MMU_NOTIFIER
int kvm_unmap_hva_range(struct kvm *kvm,
			unsigned long start, unsigned long end, bool blockable);
void kvm_set_spte_hva(struct kvm *kvm, unsigned long hva, pte_t pte);
int kvm_age_hva(struct kvm *kvm, unsigned long start, unsigned long end);
int kvm_test_age_hva(struct kvm *kvm, unsigned long hva);

static inline void kvm_get_badinstr(struct kvm_vcpu_arch *arch, unsigned int *out)
{
	*out = arch->host_badinstr;
}

static inline void update_pc(struct kvm_vcpu_arch *arch)
{
	arch->pc += 4;
}

/**
 * kvm_is_ifetch_fault() - Find whether a TLBL exception is due to ifetch fault.
 * @vcpu:	Virtual CPU.
 *
 * Returns:	Whether the TLBL exception was likely due to an instruction
 *		fetch fault rather than a data load fault.
 */
static inline bool kvm_is_ifetch_fault(struct kvm_vcpu_arch *arch)
{
	if (arch->pc == arch->host_badvaddr)
		return true;

	return false;
}


u32 kvm_loongarch_read_stable_timer(struct kvm_vcpu *vcpu);
void kvm_loongarch_init_stable_timer(struct kvm_vcpu *vcpu,
				    unsigned long stable_timer_hz);
int kvm_loongarch_set_stable_timer_ctl(struct kvm_vcpu *vcpu,
					     s64 stable_timer_ctl);
int kvm_loongarch_set_stable_timer_resume(struct kvm_vcpu *vcpu,
					  s64 stable_timer_resume);
int kvm_loongarch_set_stable_timer_hz(struct kvm_vcpu *vcpu,
					      s64 stable_timer_hz);
void kvm_loongarch_write_stable_timer(struct kvm_vcpu *vcpu, u64 stable_timer);
int kvm_loongarch_stable_timer_disabled(struct kvm_vcpu *vcpu);
int kvm_loongarch_restore_hrtimer(struct kvm_vcpu *vcpu, ktime_t before,
					     u64 stable_timer, int min_drift);
ktime_t kvm_loongarch_freeze_hrtimer(struct kvm_vcpu *vcpu, u64 *stable_timer);
enum hrtimer_restart kvm_loongarch_count_timeout(struct kvm_vcpu *vcpu);

/* fairly internal functions requiring some care to use */

void kvm_loongarch_acquire_htimer(struct kvm_vcpu *vcpu);
void kvm_loongarch_lose_htimer(struct kvm_vcpu *vcpu);

int kvm_larch_emu_st(larch_inst inst, struct kvm_vcpu *vcpu);
int kvm_larch_emu_ld(larch_inst inst, struct kvm_vcpu *vcpu);
int kvm_larch_complete_ld(struct kvm_vcpu *vcpu, struct kvm_run *run);
int kvm_loongarch_emul_wait(struct kvm_vcpu *vcpu);
int kvm_loongarch_emul_hypcall(struct kvm_vcpu *vcpu, larch_inst inst);
int kvm_loongarch_handle_hypcall(struct kvm_vcpu *vcpu);

/* Misc */
extern void kvm_loongarch_dump_stats(struct kvm_vcpu *vcpu);
extern unsigned long kvm_loongarch_get_ramsize(struct kvm *kvm);

static inline void kvm_arch_hardware_unsetup(void) {}
static inline void kvm_arch_sync_events(struct kvm *kvm) {}
static inline void kvm_arch_free_memslot(struct kvm *kvm,
		struct kvm_memory_slot *free, struct kvm_memory_slot *dont) {}
static inline void kvm_arch_memslots_updated(struct kvm *kvm, u64 gen) {}
static inline void kvm_arch_sched_in(struct kvm_vcpu *vcpu, int cpu) {}
static inline void kvm_arch_vcpu_blocking(struct kvm_vcpu *vcpu) {}
static inline void kvm_arch_vcpu_unblocking(struct kvm_vcpu *vcpu) {}
static inline void kvm_arch_vcpu_block_finish(struct kvm_vcpu *vcpu) {}

extern int kvm_enter_guest(struct kvm_run *run, struct kvm_vcpu *vcpu);
extern void kvm_exception_entry(void);
#endif /* __LOONGARCH_KVM_HOST_H__ */
